//region Helper functions
function safe_key_union(sourceA, sourceB) {
    const aBad = sourceA === undefined || sourceA === null;
    const bBad = sourceB === undefined || sourceB === null;

    if (aBad && bBad) return new Set();

    if (aBad) {
        const bKeys = Object.keys(sourceB);
        return new Set(bKeys);
    }
    const aKeys = Object.keys(sourceA);

    if (bBad) {
        return new Set(aKeys);
    }
    const bKeys = Object.keys(sourceB);

    let setA = new Set(aKeys);
    let setB = new Set(bKeys);

    for (let elem of setB) {
        setA.add(elem);
    }
    return setA;
}

//endregion

class Mask {

    constructor() {
        this.container = {};
    }

    setRed(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] |= 0x00FF0000;
    }

    setGreen(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] |= 0x0000FF00;
    }

    setBlue(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] |= 0x000000FF;
    }

    clearRed(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] &= ~0x00FF0000;
    }

    clearGreen(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] &= ~0x0000FF00;
    }

    clearBlue(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] &= ~0x000000FF;
    }

    testRed(x, y) {
        return x in this.container && y in this.container[x] && (this.container[x][y] & 0x00FF0000) !== 0;
    }

    testGreen(x, y) {
        return x in this.container && y in this.container[x] && (this.container[x][y] & 0x0000FF00) !== 0;
    }

    testBlue(x, y) {
        return x in this.container && y in this.container[x] && (this.container[x][y] & 0x000000FF) !== 0;
    }

    ensureAvailable(x, y) {
        if (!(x in this.container)) {
            this.container[x] = {};
        }
        if (!(y in this.container[x])) {
            this.container[x][y] = 0xFF000000;
        }
    }

    getMask(x, y) {
        if (!(x in this.container)) return null;
        if (!(y in this.container[x])) return null;
        return this.container[x][y];

    }

    setMask(xKey, yKey, value) {
        this.ensureAvailable(xKey, yKey);
        this.container[xKey][yKey] = value;
    }

    union(that) {
        let result = new Mask()

        let xKeys = safe_key_union(this.container, that.container);
        for (const xKey of xKeys) {
            let yKeys = safe_key_union(this.container[xKey], that.container[xKey]);
            for (const yKey of yKeys) {
                const thisValue = this.getMask(xKey, yKey);
                const thatValue = that.getMask(xKey, yKey);

                let value;
                if (thisValue === null || thisValue === undefined) {
                    value = thatValue;
                } else if (thatValue === null || thatValue === undefined) {
                    value = thisValue;
                } else {
                    value = thisValue | thatValue;
                }
                result.setMask(xKey, yKey, value);
            }
        }

        return result;
    }
}

class Rectangle {
    constructor(x, y, w, h) {
        this.x = x;
        this.y = y;
        this.w = w;
        this.h = h;
    }

    coordinateInside(x, y) {
        return !(x < this.x || y < this.y || x >= this.x + this.w || y >= this.y + this.h);
    }

    area() {
        return this.w * this.h;
    }

    scale(w_scalar, h_scalar) {
        const width_increase = Math.ceil(this.w * w_scalar)-this.w;
        const height_increase = Math.ceil(this.h * h_scalar)-this.h;

        this.x -= Math.ceil(width_increase / 2.0);
        this.y -= Math.ceil(height_increase / 2.0);

        this.w +=  width_increase;
        this.h +=  height_increase;
    }

    scaled(w_scalar, h_scalar) {
        const new_rect = this.clone();
        new_rect.scale(w_scalar, h_scalar);
        return new_rect;
    }

    stretch(width, height) {
        this.x -= width;
        this.w += 2 * width;

        this.y -= height;
        this.h += 2 * height;
    }

    stretched(width, height) {
        const new_rect = this.clone();
        new_rect.stretch(width, height);
        return new_rect;
    }

    clone() {
        return new Rectangle(this.x, this.y, this.w, this.h);
    }
}

class GlitchConfig {
    constructor() {
        this.r_channel_intensity = 0;
        this.g_channel_intensity = 0;
        this.b_channel_intensity = 0;
        this.r_negate_intensity = 0;
        this.g_negate_intensity = 0;
        this.b_negate_intensity = 0;
        this.shift_horizontal = false;
        this.shift_vertical = false;
        this.h_bar_size_range = [0, 0];
        this.v_bar_size_range = [0, 0];
        this.shift_intensity_range = [0, 0];
    }

    set_channel_intensities(value) {
        this.r_channel_intensity = value;
        this.g_channel_intensity = value;
        this.b_channel_intensity = value;
    }

    set_negation_intensities(value) {
        this.r_negate_intensity = value;
        this.g_negate_intensity = value;
        this.b_negate_intensity = value;
    }

    set_all_intensities(value) {
        this.set_channel_intensities(value);
        this.set_negation_intensities(value);
    }
}

class ImageProcessor {
    constructor(original, bar_configuration, pixel_configuration, blur_configuration, glitch_configuration) {
        // Cloing mavinimages takes 25 to 600 ms! and is only needed for glitch mode right now


        this.marvin_out = original

        this.bar_configuration = bar_configuration;
        this.pixel_configuration = pixel_configuration;
        this.blur_configuration = blur_configuration;
        this.glitch_configuration = glitch_configuration;

        this.natural_width = original.canvas.width;
        this.natural_height = original.canvas.height;

        this.blurred = null;
        this.marvin_original = null;
        // this.whole_image_adjust = false;

        this.per_layer_queues = []
    }

    pixelateArea(rect) {
        const pixel_configuration = this.pixel_configuration;
        rect = rect.scaled(pixel_configuration._scale, pixel_configuration._scale);

        let distance = pixel_configuration._distance;
        let box_size = pixel_configuration._pixel_size;
        // TODO fix _geometry to be an int (or other non magic-number value)
        let geometry = parseInt(pixel_configuration._geometry);

        if (pixel_configuration._pixel_size_mode === 0) {
            const smallest_dimension = Math.min(this.natural_height, this.natural_width);
            box_size = Math.ceil(smallest_dimension / 100 * pixel_configuration._pixel_size);
        }

        // Hexelate
        if (geometry === 1) {
            hexelate_area(this.marvin_out, rect, box_size, distance, false);

            // Hexelate glitch
        } else if (geometry === 2) {
            hexelate_area(this.marvin_out, rect, box_size, distance, true);

        } else {
            pixelate_rectangle(this.marvin_out, rect, box_size, distance);
            /*
            if (box_size < 5) {
                pixelate_rectangle_fast(this.marvin_out, rect, box_size, distance);
            } else {
                pixelate_rectangle(this.marvin_out, rect, box_size, distance);
            }
            */
        }
    }

    blurArea(rect) {
        const blur_configuration = this.blur_configuration;
        if (this.blurred == null) {
            this.blurred = generateBaseBlurImproved(this.marvin_out, blur_configuration._blur_strength);
        }
        this.marvin_out.update();

        let canvas = document.createElement('canvas');
        let original_w = rect.w;
        let original_h = rect.h;
        rect = rect.scaled(blur_configuration._scale, blur_configuration._scale);
        canvas.width = rect.w;
        canvas.height = rect.h;

        const ctx = canvas.getContext('2d');

        ctx.fillStyle = 'rgba(0,0,0,0)';
        ctx.fillRect(0, 0, rect.w, rect.h);
        ctx.fillStyle = '#fff';
        let blur_radius_strength = Math.ceil(Math.min(rect.w, rect.h) / 100 * blur_configuration._blur_radius);
        ctx.filter = 'blur(' + blur_radius_strength + 'px)';
        ctx.beginPath();
        ctx.ellipse(rect.w / 2, rect.h / 2,
            original_w*blur_configuration._mask_scale,
            original_h*blur_configuration._mask_scale, 0, 0, 4 * Math.PI);
        ctx.fill();
        ctx.globalCompositeOperation = 'source-in';
        ctx.drawImage(this.blurred, rect.x, rect.y, rect.w, rect.h, 0, 0, rect.w, rect.h);

        this.marvin_out.ctx.drawImage(canvas, 0, 0, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
        //marvinImage.ctx.clearRect(0, 0, marvinImage.width,marvinImage.height);
        //marvinImage.ctx.drawImage(marvinImage.blurred, 0,0,marvinImage.width,marvinImage.height);
        this.marvin_out.imageData = this.marvin_out.ctx.getImageData(0, 0, this.marvin_out.width, this.marvin_out.height);
    }

    blackenArea(rect) {
        const bar_configuration = this.bar_configuration;
        //this.marvin_out.fillRect(rect.x, rect.y, rect.w, rect.h, bar_configuration._color);
        this.marvin_out.ctx.fillStyle = bar_configuration._color;
        rect = rect.scaled(bar_configuration._scale, bar_configuration._scale);
        this.marvin_out.ctx.fillRect(rect.x, rect.y, rect.w, rect.h);
        this.marvin_out.imageData =  this.marvin_out.ctx.getImageData(0, 0, this.natural_width, this.natural_height);
    }

    drawTextBox(ctx, x, y, text, bg_color, baseline = "top", font_sz = 16, font_family = "sans-serif", fg_color = "#000000") {
        ctx.font = `${font_sz}px ${font_family}`;
        ctx.textBaseline = baseline;

        const metrics = ctx.measureText(text);
        const string_width = Math.ceil(ctx.measureText(text).width);
        const font_height = font_sz

        let y_adjust = baseline === "bottom" ? -(2 + font_height) : 0;
        ctx.fillStyle = bg_color;
        ctx.fillRect(x, y + y_adjust, string_width + 4, font_height + 4);

        // text = text + `sz: ${string_width}`;
        ctx.fillStyle = fg_color;
        ctx.fillText(text, x, y);
    }

    namedBoxRect(klass, score, rect) {
        const bg_color = "#00FFFF";

        let font_size = 2 + Math.ceil((this.natural_width * 14) / 1000);
        font_size = Math.max(9, font_size);

        const ctx = this.marvin_out.ctx;

        ctx.strokeStyle = bg_color;
        ctx.lineWidth = 4;

        ctx.strokeRect(rect.x, rect.y, rect.w, rect.h);

        let text_to_write = klass.name;
        this.drawTextBox(ctx, rect.x, rect.y, text_to_write, bg_color, "top", font_size);

        text_to_write = score;
        this.drawTextBox(ctx, rect.x, rect.y + rect.h, text_to_write, bg_color, "bottom", font_size - 2);

        this.marvin_out.imageData = ctx.getImageData(0, 0, this.natural_width, this.natural_height);
    }

    fettyfiArea(klass, rect) {
        if(this.marvin_original == null) {
            //let t0 = performance.now();
            this.marvin_original = this.marvin_out.clone();
            //let t1 = performance.now();
            //console.log("clone marvin_original time " + (t1 - t0) + " millisseconds.");
        }
        let mul = 1;
        if (klass.erotic) {
            mul = 5;
        }
        mul = Math.ceil(mul*this.glitch_configuration.intensity);

        let config;

        let canvas = this.marvin_out.clone();
        let source = this.marvin_original;
        rect = rect.scaled(this.glitch_configuration._scale, this.glitch_configuration._scale);
        let effect_range = Math.ceil(0.3 * Math.min(rect.h, rect.w));
        let effect_bounds = rect.stretched(effect_range, effect_range);

        // Round 1:
        // Green, Horizontal, 50%, default bar & shift sizes
        config = new GlitchConfig();
        config.g_channel_intensity = 0.75 * mul;
        // config.set_channel_intensities(0.05 * mul);
        config.shift_horizontal = this.glitch_configuration.shift_horizontal;
        config.h_bar_size_range = [Math.floor(rect.h / 20), Math.ceil(rect.h / 10)];
        config.shift_intensity_range = [Math.round(effect_range / 4), effect_range];

        let mask1 = this.neo_shift(source, canvas, rect, effect_bounds, config)
        source = canvas.clone();

        // Round 2:
        // Red, Vertical, 200%, default bar & shift sizes.
        config = new GlitchConfig();
        config.r_channel_intensity = 0.85 * (1 + mul);
        // config.set_channel_intensities(0.6 * mul);
        config.shift_vertical = this.glitch_configuration.shift_vertical;
        config.v_bar_size_range = [Math.floor(rect.w / 20), Math.ceil(rect.w / 10)];
        config.shift_intensity_range = [Math.round(effect_range / 12), Math.round(effect_range / 2)];

        let mask2 = this.neo_shift(source, canvas, rect, effect_bounds, config)
        let mask3 = mask1.union(mask2);

        let area = effect_bounds.area();
        let logSize = Math.log(area);
        let sizeScalar = Math.round(logSize);
        //console.warn("Size Scalar for ", klass.name, " is ", sizeScalar);

        this.enqueue(20 - sizeScalar, effect_bounds, canvas, mask3);
    }

    enqueue(layer, rect, intense_canvas, mask) {
        if (!(layer in this.per_layer_queues)) {
            this.per_layer_queues[layer] = [];
        }
        this.per_layer_queues[layer].push([rect, intense_canvas, mask]);
    }

    applyEffect(klass, score, rect, type) {
        let t0 = performance.now();
        switch (type) {
            case effects.PIXEL:
                this.pixelateArea(rect);
                break;
            case effects.BLUR:
                this.blurArea(rect);
                break;
            case effects.BLACK:
                this.blackenArea(rect);
                break;
            case effects.BOX:
                this.namedBoxRect(klass, score, rect);
                break;
            case effects.NONE:
                break;
            case effects.GLITCH:
                this.fettyfiArea(klass, rect);
                break;
        }
        let t1 = performance.now();
        log("Effect processing took " + (t1 - t0) + " ms.");
        // return this.marvin_out
    }

    finalize() {
        for (let layer_index = 0; layer_index < this.per_layer_queues.length; layer_index++) {
            if (layer_index in this.per_layer_queues) {

                let entries = this.per_layer_queues[layer_index];

                entries.forEach((entry) => {
                    // console.warn("You were warned!", JSON.stringify(entry, null, 4));
                    let rect = entry[0];
                    let canvas = entry[1];
                    let mask = entry[2];
                    this.apply_section(rect, canvas, mask);
                });

            }
        }
        this.marvin_out.update();
    }

    apply_section(rect, canvas, mask) {
        for (let x in mask.container) {
            if (!mask.container.hasOwnProperty(x)) continue;
            x = parseInt(x);

            for (let y in mask.container[x]) {
                if (!mask.container[x].hasOwnProperty(y)) continue;
                y = parseInt(y);

                const coordinate_mask = mask.container[x][y];
                const fg_color = canvas.getIntColor(x, y) & coordinate_mask;
                const bg_color = this.marvin_out.getIntColor(x, y) & ~coordinate_mask;

                this.marvin_out.setIntColor1(x, y, fg_color | bg_color);
                // this.marvin_out.setIntColor1(x, y, coordinate_mask);
            }
        }
    }

    neo_shift(source, canvas, rect, effect_bounds, config) {
        // TODO implement internal H/V distinction.

        let shift_range = config.shift_intensity_range;
        // console.warn("h calc: ", rect);
        const hShifts = this.calculateShifts(rect.h, config.h_bar_size_range, shift_range);
        // console.warn("v calc: ");
        const vShifts = this.calculateShifts(rect.w, config.v_bar_size_range, shift_range);

        let mask = new Mask();
        for (let dy = 0; dy < rect.h; dy++) {
            const y = rect.y + dy;
            const hShift = hShifts[dy];

            for (let dx = 0; dx < rect.w; dx++) {
                const x = rect.x + dx;

                // const a = source.getAlphaComponent(x, y);
                const r = Math.round(source.getIntComponent0(x, y) * config.r_channel_intensity);
                const g = Math.round(source.getIntComponent1(x, y) * config.g_channel_intensity);
                const b = Math.round(source.getIntComponent2(x, y) * config.b_channel_intensity);

                const vShift = vShifts[dx];
                const tx = x + hShift;
                const ty = y + vShift;

                const ta = source.getAlphaComponent(tx, ty)
                const tr = Math.round(source.getIntComponent0(tx, ty) * (1 - config.r_channel_intensity));
                const tg = Math.round(source.getIntComponent1(tx, ty) * (1 - config.g_channel_intensity));
                const tb = Math.round(source.getIntComponent2(tx, ty) * (1 - config.b_channel_intensity));

                if (!effect_bounds.coordinateInside(tx, ty)) continue;
                if (tx < 0 || tx >= canvas.getWidth() || ty < 0 || ty >= canvas.getHeight()) continue;

                if (config.r_channel_intensity > 0) mask.setRed(tx, ty);
                if (config.g_channel_intensity > 0) mask.setGreen(tx, ty);
                if (config.b_channel_intensity > 0) mask.setBlue(tx, ty);
                canvas.setIntColor(tx, ty, ta, tr + r, tg + g, tb + b);

                // const ultimate_combined_pixel = calculate_combined_pixel(original, x_global, y_global, shift, 0, config);
                // local_image.setIntColor(x_global, y_global, ultimate_combined_pixel);
            }
        }
        return mask;
    }

    calculateShifts(dimension, bar_range, shift_range) {
        // console.warn("Processing shift calculation for: ", dimension, bar_range, shift_range);
        if (bar_range[1] === 0 && bar_range[0] === 0) return new Array(dimension).fill(0);
        let bar_size = 0;
        let shift_intensity = 0;
        const indexed_shifts = [];
        for (let i = 0; i < dimension; i++) {
            if (bar_size > 0) {
                bar_size--;
            } else {
                if (dimension - i < bar_range[1]) break;
                bar_size = generate_bar_size(bar_range[0], bar_range[1]);
                shift_intensity = generate_shift_intensity(shift_range[0], shift_range[1], shift_intensity);
            }
            indexed_shifts[i] = shift_intensity;
        }
        return indexed_shifts;
    }
}

function int_to_argb(int_color) {
    const a_channel = (int_color >> 24) & 0xFF;
    const r_channel = (int_color >> 16) & 0xFF;
    const g_channel = (int_color >> 8) & 0xFF;
    const b_channel = int_color & 0xFF;
    return [a_channel, r_channel, g_channel, b_channel]
}

function argb_to_int(channels) {
    const a_channel = (channels[0] & 0xFF) << 24;
    const r_channel = (channels[1] & 0xFF) << 16;
    const g_channel = (channels[2] & 0xFF) << 8;
    const b_channel = (channels[3] & 0xFF);
    return a_channel | r_channel | g_channel | b_channel;
}

function generate_bar_size(min, max) {
    const random_value = Math.random() * (max - min);
    return Math.floor(min + random_value);
}

function generate_shift_intensity(min, max, previous) {
    const r_base = min + Math.round(Math.random() * (max - min));
    const r_sign = Math.round(Math.random()) === 0 ? -1 : 1;
    const random_delta = r_base * r_sign;

    let candidate_shift = previous + random_delta;
    if (candidate_shift < -max || max < candidate_shift) {
        candidate_shift = previous - random_delta;
    }
    return candidate_shift;
}

function make_region_buffer(local_image, x_origin, y_origin, width, height) {
    const buffer = new Array(width * height);

    for (let y_local = 0; y_local < height; y_local++) {
        let y_global = y_origin + y_local;
        for (let x_local = 0; x_local < width; x_local++) {
            let x_global = x_origin + x_local;
            buffer[y_local * width + x_local] = local_image.getIntColor(x_global, y_global)
        }
    }
    return buffer;
}

function calculate_combined_pixel(original, x_global, y_global, x_shift, y_shift, config) {
    const original_pixel = original.getIntColor(x_global, y_global);
    const original_argb = int_to_argb(original_pixel);

    const shifted_pixel = original.getIntColor(x_global + x_shift, y_global + y_shift)
    const shifted_argb = int_to_argb(shifted_pixel);

    // noinspection DuplicatedCode
    return actually_combine_pixel_values(original_argb, shifted_argb, config);
}

function actually_combine_pixel_values(original_argb, shifted_argb, config) {
    const combined_argb = [
        original_argb[0],
        shifted_argb[1] * config.r_channel_intensity + original_argb[1] * (1 - config.r_channel_intensity),
        shifted_argb[2] * config.g_channel_intensity + original_argb[2] * (1 - config.g_channel_intensity),
        shifted_argb[3] * config.b_channel_intensity + original_argb[3] * (1 - config.b_channel_intensity),
    ];

    // noinspection DuplicatedCode
    const negative_combined_argb = [
        original_argb[0],
        (0xFF - shifted_argb[1]) * config.r_channel_intensity + (0xFF - original_argb[1]) * (1 - config.r_channel_intensity),
        (0xFF - shifted_argb[2]) * config.g_channel_intensity + (0xFF - original_argb[2]) * (1 - config.g_channel_intensity),
        (0xFF - shifted_argb[3]) * config.b_channel_intensity + (0xFF - original_argb[3]) * (1 - config.b_channel_intensity),
    ];

    // noinspection DuplicatedCode
    const ultimate_combined_argb = [
        combined_argb[0],
        negative_combined_argb[1] * config.r_negate_intensity + combined_argb[1] * (1 - config.r_negate_intensity),
        negative_combined_argb[2] * config.g_negate_intensity + combined_argb[2] * (1 - config.g_negate_intensity),
        negative_combined_argb[3] * config.b_negate_intensity + combined_argb[3] * (1 - config.b_negate_intensity),
    ];

    return argb_to_int(ultimate_combined_argb);
}

function merge_pixels(bg_pixel, fg_pixel) {

    const fg_argb = int_to_argb(fg_pixel);
    const bg_argb = int_to_argb(bg_pixel);

    const fg_factor = (fg_argb[0] / 255.);
    const bg_factor = 1 - fg_factor;

    const combined_argb = [
        bg_argb[0],
        fg_argb[1] & 0b1 ? bg_argb[1] * bg_factor + fg_argb[1] * fg_factor : bg_argb[1],
        fg_argb[2] & 0b1 ? bg_argb[2] * bg_factor + fg_argb[2] * fg_factor : bg_argb[2],
        fg_argb[3] & 0b1 ? bg_argb[3] * bg_factor + fg_argb[3] * fg_factor : bg_argb[3],
    ];

    return argb_to_int(combined_argb);
}

function generateBaseBlurImproved(marvinImage, blur_strength) {
    marvinImage.update();
    let canvas = document.createElement('canvas');
    let pxext = blur_strength * 2;
    let pxext2 = pxext * 2;
    canvas.width = marvinImage.width + pxext2;
    canvas.height = marvinImage.height + pxext2;
    const ctx = canvas.getContext('2d');
    ctx.save();
    //Upper left corner
    ctx.translate(pxext, pxext);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, 0, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Top
    ctx.translate(0, pxext);
    ctx.scale(1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, 0, marvinImage.width, pxext,
        pxext, 0, marvinImage.width, pxext);
    ctx.restore();
    ctx.save();
    //Upper right corner
    ctx.translate(canvas.width, pxext);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        marvinImage.width - pxext, 0, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Left
    ctx.translate(pxext, 0);
    ctx.scale(-1, 1);
    ctx.drawImage(marvinImage.canvas,
        0, 0, pxext, marvinImage.height,
        0, pxext, pxext, marvinImage.height
        //0, 0, marvinImage.width, marvinImage.height
    );
    ctx.restore();
    ctx.save();
    //Right
    ctx.translate(canvas.width, 0);
    ctx.scale(-1, 1);
    ctx.drawImage(marvinImage.canvas,
        marvinImage.width - pxext, 0, pxext, marvinImage.height,
        0, pxext, pxext, marvinImage.height);
    ctx.restore();
    ctx.save();
    //Bottom left corner
    ctx.translate(pxext, canvas.height);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, marvinImage.height - pxext, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Bottom
    ctx.translate(0, canvas.height);
    ctx.scale(1, -1);
    ctx.drawImage(marvinImage.canvas,
        0, marvinImage.height - pxext, marvinImage.width, pxext,
        pxext, 0, marvinImage.width, pxext);
    ctx.restore();
    ctx.save();
    //Bottom right corner
    ctx.translate(canvas.width, canvas.height);
    ctx.scale(-1, -1);
    ctx.drawImage(marvinImage.canvas,
        marvinImage.width - pxext, marvinImage.height - pxext, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Center
    ctx.drawImage(marvinImage.canvas, pxext, pxext, marvinImage.width, marvinImage.height);
    ctx.filter = 'blur(' + blur_strength + 'px)';
    ctx.drawImage(canvas, 0, 0, canvas.width, canvas.height);
    let blurred = document.createElement('canvas');
    blurred.width = marvinImage.width;
    blurred.height = marvinImage.height;
    const ctx_blurred = blurred.getContext('2d');

    ctx_blurred.drawImage(canvas,
        pxext, pxext, marvinImage.width, marvinImage.height,
        0, 0, marvinImage.width, marvinImage.height);

    return blurred;
}

function pixelate_rectangle(marvinImage, rect, pixel_size = 20, distance = 0) {
    marvinImage.update();
    let pixelArr = marvinImage.ctx.getImageData(0, 0, marvinImage.getWidth(), marvinImage.getHeight()).data;
    for (let i = rect.x; i < rect.x + rect.w; i += pixel_size + distance) {
        for (let j = rect.y; j < rect.y + rect.h; j += pixel_size + distance) {
            marvinImage.ctx.fillStyle = calculatePixelBoxColor(marvinImage, i, j, i + pixel_size, j + pixel_size, pixel_size, pixelArr);
            marvinImage.ctx.fillRect(i, j, pixel_size, pixel_size);
        }
    }
    marvinImage.imageData =  marvinImage.ctx.getImageData(0, 0, marvinImage.getWidth(), marvinImage.getHeight());
    return marvinImage;
}

function toColor(num, alpha = false) {
    num >>>= 0;
    let b = num & 0xFF,
        g = (num & 0xFF00) >>> 8,
        r = (num & 0xFF0000) >>> 16,
        a = ((num & 0xFF000000) >>> 24) / 255;
    if (!alpha) {
        a = 1;
    }
    return "rgba(" + [r, g, b, a].join(",") + ")";
}

function hexelate_area(marvinImage, rect, pixel_size = 20, distance = 0, glitch = false) {
    let a = 2 * Math.PI / 6;
    let r = pixel_size;
    if (glitch) {
        a = Math.ceil(a);
    }

    function drawHexagon(marvinImage, x, y, r, a, pixelArr) {

        marvinImage.ctx.beginPath();
        for (let i = 0; i < 6; i++) {
            marvinImage.ctx.lineTo(x + r * Math.cos(a * i), y + r * Math.sin(a * i));
        }
        marvinImage.ctx.fillStyle = calculatePixelBoxColor(marvinImage, Math.ceil(x), Math.ceil(y), 0, 0, r, pixelArr);

        marvinImage.ctx.fill();
        marvinImage.ctx.closePath();
    }
    marvinImage.update();
    let pixelArr = marvinImage.ctx.getImageData(0, 0, marvinImage.getWidth(), marvinImage.getHeight()).data;

    for (let y = rect.y + r; y + r * Math.sin(a) < rect.y + rect.h; y += 2 * (r * Math.sin(a))) {
        y += distance;
        for (let x = rect.x + r, j = 0; x + r * (1 + Math.cos(a)) < rect.x + rect.w; x += r * (1 + Math.cos(a))) {
            x += distance;
            let yc = y + 0.5 * ((-1) ** j++ * r * Math.sin(a));
            drawHexagon(marvinImage, x, yc, r, a, pixelArr);
        }
    }
    marvinImage.imageData = marvinImage.ctx.getImageData(0, 0, marvinImage.width, marvinImage.height);
    marvinImage.update();
    return marvinImage;
}

/**
 *
 * @param marvinImage
 * @param x1 = pixel box starting x
 * @param y1 = pixel box starting y
 * @param x2 = pixel box ending x
 * @param y2 = pixel box ending y
 * @param pixel_size = box size
 * @param pixelArr = pixel array
 * @returns {*}
 */
function calculatePixelBoxColor(marvinImage, x1, y1, x2, y2, pixel_size, pixelArr) {

    let p = (x1 + (y1*marvinImage.getWidth())) * 4;
    return "rgba(" + pixelArr[p] + "," + pixelArr[p + 1] + "," + pixelArr[p + 2] + "," + pixelArr[p + 3] + ")";
    /* These had a bad performance
    switch (1) {
        case 0:
            return marvinImage.getIntColor(x1, y1);
        case 1:
            let rectarea = new MarvinImage();
            let pixelmarvin = new MarvinImage();
            Marvin.crop(marvinImage, rectarea, x1, y1, pixel_size, pixel_size);
            Marvin.scale(rectarea, pixelmarvin, 1, 1);
            return pixelmarvin.getIntColor(0, 0);
        default:
    }
    return marvinImage.getIntColor(x1, y1);
    */
}